#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <locale.h>
#include <zlib.h>

#include "xp3ext.h"

static struct {
  FILE*              fp;
  struct xp3head     head;
  struct xp3filehead filehead;
  int                filenum;
  struct xp3fileinfo *fileinfos;
} xp3dat;

#define DEBUG 0

// fobOo
void  debug(char *format, ...)
{
  va_list ap;

  if (!DEBUG)
    return;
  fprintf(stderr, "Debug: ");
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  fprintf(stderr, "\n");
}

// G[bZ[W\ďI
void  error_exit(char *format, ...)
{
  va_list ap;

  fprintf(stderr, "Error: ");
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  fprintf(stderr, "\n");
  exit(1);
}


// mۂ֐BG[`FbNtB
void *alloc(size_t siz)
{
  void *ret;
  if ((ret = malloc(siz)) == NULL)
    error_exit("alloc(%lld(=0x%08llx)) in %s()", siz, siz, __FUNCTION__);
  return ret;
}


// kf[^AzlibgēWJ
void  decode(void *dst, uint64_t dstsiz, void *src, uint64_t srcsiz)
{
  int err;
  unsigned long siz;

  if ((err = uncompress(dst, &siz, src, srcsiz)) != Z_OK || dstsiz != siz)
    error_exit("error code = %d, Fail to uncompress()", err);
}


#if 0
// wchar_t*  char* ɕϊBLinuxwchar_tƈႤ̂Ńc
int  wcstombs_local(char *cs, wchar_t *wcs, size_t n)
{
  int i, idx;
  uint8_t *dst = (uint8_t*)cs, *src = (uint8_t*)wcs;

  for (i = 0, idx = 0; idx < n; i+=2) {
    dst[idx++] = src[i];
    if (src[i] == 0)
      break;
    // Shift-JISEUCO񂾂ǂ납BUTF-32̂ƂƂȂ납B
    if (src[i] >= 0x80)
      dst[idx++] = src[i+1];
  }
  return idx;
}
#endif


// t@C쐬邽߂ɕKvȃfBNg
void  mkdir_p(char *fnam)
{
  int  i;
  char dirnam[PATH_MAX+1];

  for (i = 1/*0͕sv*/; i < strlen(fnam); i++) {
    if (fnam[i] == '/') {
      strncpy(dirnam, fnam, i);
      dirnam[i] = 0;
      if (mkdir(dirnam, 0755) != 0 && errno != EEXIST)
	error_exit("fail to mkdir(%s) in %s(%s)", dirnam, __FUNCTION__, fnam);
    }
  }
}


// xp3t@C̃t@Cwb_ǂݍ݁Axp3dat.head Ɋi[
void  read_xp3header(void)
{
  int8_t  xp3magic[] = XP3MAGIC;
  int64_t tmpidx;

  debug("Enter %s()", __FUNCTION__);

  // }WbNNo.`FbN
  fidx(xp3dat.fp, 0);
  fload(xp3dat.fp, xp3dat.head.magic, sizeof(xp3dat.head.magic));
  if (memcmp(xp3dat.head.magic, xp3magic, sizeof(xp3magic)) != 0)
    error_exit("xp3magic is mis-match in %s()", __FUNCTION__);

  // 64bitlŁA2.28ȑO2.30ȍ~𔻕ʁB̂Ȃ̂ŁB
  fint64(xp3dat.fp, tmpidx);
  if (tmpidx != CUSHIONIDX) {
    xp3dat.head.fileinfo_idx = tmpidx;
    // ̂ƂAI xp3dat.head.minor_version = 0 ɂȂB
  } else {
    fint32(xp3dat.fp, xp3dat.head.minor_version);
    fint8 (xp3dat.fp, xp3dat.head.cushion_head);
    fint64(xp3dat.fp, xp3dat.head.cushion_idx_siz);

    fint64(xp3dat.fp, xp3dat.head.fileinfo_idx);
  }
  debug("fileinfo_idx = %08llx", xp3dat.head.fileinfo_idx);

  debug("Exit %s()", __FUNCTION__);
}


// ZOgǂ݂
struct xp3segment  read_a_segment(void *buf)
{
  struct xp3segment seg;
  mint32(buf, seg.flag);
  mint64(buf, seg.offset);
  mint64(buf, seg.orgsiz);
  mint64(buf, seg.datsiz);
  return seg;
}


// fileinfoǂݍ
struct xp3fileinfo  read_a_fileinfo(void *buf, uint64_t bufsiz)
{
  struct xp3fileinfo fi;
  void *buftop = buf;
  int processed = 0;

  debug("Enter %s(buf, 0x%08llx)", __FUNCTION__, bufsiz);

  while (buf - buftop < bufsiz) {
    uint8_t  id[4];
    uint64_t siz;
    void     *bufsave = buf;

    // `Nwbhǂݍ
    mstrid(buf, id);
    mint64(buf, siz);

    // IDƂɏU蕪
    if (strncmp(id, FILEID, sizeof(FILEID)-1) == 0) {
      // "File"`N
      if (processed)
	break; // "File" xςȂȀɉ
      debug("Process \"File\" chank(idx=%d)", buf-buftop-CHUNKHEADSIZ);
      processed = 1;
      strncpy(fi.fileid, id, sizeof(FILEID));
      fi.siz = siz;
    }
    else if (strncmp(id, INFOID, sizeof(INFOID)-1) == 0) {
      // "info"`N
      debug("Process \"info\" chank(idx=%d)", buf-buftop-CHUNKHEADSIZ);
      strncpy(fi.infoid, id, sizeof(INFOID));
      fi.infosiz = siz;

      mint32(buf, fi.flag);
      mint64(buf, fi.orgsiz);
      mint64(buf, fi.datsiz);

      // t@C(Wstring)ۑ
      mint16(buf, fi.filenamsiz);
      fi.filenam = alloc(fi.filenamsiz*2 + 2); // I[̂߂ +2 ͕Kv
      memcpy(fi.filenam, buf, fi.filenamsiz*2);
      ((uint8_t*)fi.filenam)[fi.filenamsiz*2]   = '\0'; // ÔߏI[
      ((uint8_t*)fi.filenam)[fi.filenamsiz*2+1] = '\0';

      buf = bufsave + siz + CHUNKHEADSIZ;
    }
    else if (strncmp(id, SEGID, sizeof(SEGID)-1) == 0) {
      // "segm" `N
      int i;
      debug("Process \"segm\" chank(idx=%d)", buf-buftop-CHUNKHEADSIZ);
      strncpy(fi.segid, id, sizeof(SEGID));
      fi.segsiz = siz;
      // ZOgǂݍ
      fi.segary = alloc(fi.segsiz);
      debug("segment number = %d", fi.segsiz/SEGMENTSIZ);

      for (i = 0; i < fi.segsiz/SEGMENTSIZ; i++) {
	fi.segary[i] = read_a_segment(buf);
	buf += SEGMENTSIZ;
      }

      buf = bufsave + siz + CHUNKHEADSIZ;
    }
    else if (strncmp(id, ADLRID, sizeof(ADLRID)-1) == 0) {
      // "adlr" `N
      // ݂̓`FbNȂ̂ŖقăXLbv
      buf = bufsave + siz + CHUNKHEADSIZ;
    }
    else {
      // mȂ`Nꍇ̓XLbv
      fprintf(stderr, "Unknown id(\"%.4s\"), skip(idx=%d)\n", id, buf-buftop-CHUNKHEADSIZ);
      buf = bufsave + siz + CHUNKHEADSIZ;
    }
  }

  debug("Exit %s()", __FUNCTION__);
  return fi;
}


// t@CǗǂݍ
void  read_fileinfos(void)
{
  int  i;
  void *buf, *bufsave;
  struct xp3filehead *fhp = &xp3dat.filehead;

  debug("Enter %s()", __FUNCTION__);

  fidx(xp3dat.fp, xp3dat.head.fileinfo_idx);
  fint8(xp3dat.fp, fhp->flag);

  // ktO`FbN
  if (!is_compressed(*fhp)) {
    // 񈳏k̏ꍇ
    fint64(xp3dat.fp, fhp->orgsiz);
    buf = bufsave = alloc(fhp->orgsiz);
    fload(xp3dat.fp, buf, fhp->orgsiz);
  } else {
    // k̏ꍇ
    void *cmpbuf;
    fint64(xp3dat.fp, fhp->datsiz);
    fint64(xp3dat.fp, fhp->orgsiz);
    buf = bufsave = alloc(fhp->orgsiz);
    cmpbuf = alloc(fhp->datsiz);
    // zlibWJ
    fload(xp3dat.fp, cmpbuf, fhp->datsiz);
    decode(buf, fhp->orgsiz, cmpbuf, fhp->datsiz);
    free(cmpbuf);
  }
  // ̎_ buf ɓWJς݊Ǘi[Ă
  // t@CǗÂǂݍ
  xp3dat.fileinfos = NULL;
  for (i = 0; buf-bufsave < fhp->orgsiz; i++) {
    struct xp3fileinfo fi;
    fi = read_a_fileinfo(buf, fhp->orgsiz - (buf-bufsave));
    xp3dat.fileinfos = realloc(xp3dat.fileinfos, sizeof(fi)*(i+1));
    if (xp3dat.fileinfos == NULL)
      error_exit("realloc fileinfo in %s", __FUNCTION__);
    xp3dat.fileinfos[i] = fi;
    buf += CHUNKHEADSIZ + fi.siz;
  }
  xp3dat.filenum = i;
  free(bufsave);

  debug("Exit %s()", __FUNCTION__);
}
  

// ZOgt@CɓWJ
void  extract_a_segment(struct xp3segment seg, FILE *fp)
{
  void  *dst;

  debug("Enter %s()", __FUNCTION__);

  fidx(xp3dat.fp, seg.offset);
  dst = alloc(seg.orgsiz);
  if (!is_compressed(seg)) {
    // ZOg񈳏kꍇ
    fload(xp3dat.fp, dst, seg.orgsiz);
  } else {
    // ZOgkĂꍇ
    void *src;
    src = alloc(seg.datsiz);
    fload(xp3dat.fp, src, seg.datsiz);
    decode(dst, seg.orgsiz, src, seg.datsiz);
    free(src);
  }
  fsave(fp, dst, seg.orgsiz);
  free(dst);

  debug("Exit %s()", __FUNCTION__);
}


// t@CWJ
void  extract_a_file(struct xp3fileinfo fi)
{
  FILE *fp;
  char fnam[FILENAME_MAX+1];
  int  fnamsiz, i;

  debug("Enter %s()", __FUNCTION__);

  // t@CWstring當֕
  if ((fnamsiz = wcstombs(fnam, fi.filenam, FILENAME_MAX)) <= 0)
    error_exit("in wcstombs(,,%d) in %s()", fi.filenamsiz, __FUNCTION__);
  fnam[fnamsiz] = '\0';
  printf("Extracting file \"%s\"\n", fnam);

  // veNgĂĂj炵
  if (is_protected(fi))
    printf("This file is protected, though I do not care:-9\n");

  // t@CWJ
  mkdir_p(fnam);
  if ((fp = fopen(fnam, "w")) == NULL) {
    // sꍇ̓XLbv邾
    fprintf(stderr, "Skip: fail to fopen(\"%s\", \"w\") in %s()", fnam, __FUNCTION__);
  } else {
    for (i = 0; i < fi.segsiz/SEGMENTSIZ; i++)
      extract_a_segment(fi.segary[i], fp);
    fclose(fp);
  }

  debug("Exit %s()", __FUNCTION__);
}


// t@CQWJ
void  extract_files(void)
{
  int  i;

  debug("Enter %s()", __FUNCTION__);

  for (i = 0; i < xp3dat.filenum; i++) {
    extract_a_file(xp3dat.fileinfos[i]);
  }

  debug("Exit %s()", __FUNCTION__);
}


// C[`
int  main(int argc, char *argv[])
{
  printf("XP3 extractor [%s] version %s %s for Linux/Cygwin by KAICHO\n", argv[0], XP3EXT_VERSION, XP3EXT_DATE);
  if (argc != 2)
    error_exit("argument number\nxp3ext <.xp3file>");

  // t@CSJISɂ邽߂ɂȂƂĂ܂B̂B
  setlocale(LC_ALL, "ja_JP.SJIS");

  // ܂f[^SĂNA
  memset(&xp3dat, 0, sizeof(xp3dat.head));

  // t@CI[v
  if ((xp3dat.fp = fopen(argv[1], "r")) == NULL)
    error_exit("open file(%s)", argv[1]);

  // wb_ǂݍ
  read_xp3header();
  // t@Cǂݍ
  read_fileinfos();

  // t@CWJ
  extract_files();

  fclose(xp3dat.fp);
  return 0;
}

